import torch
from jutility import cli
from juml.datasets.split import DataSplit
from juml.datasets.synthetic import Synthetic
from juml.loss.base import Loss
from juml.loss.mse import Mse
from juml.loss.crossentropy import CrossEntropy

class RandomClassification(Synthetic):
    def __init__(
        self,
        input_shape:    list[int],
        output_shape:   list[int],
        train:          int,
        test:           int,
    ):
        self._init_synthetic(
            input_shape=input_shape,
            output_shape=output_shape,
            n_train=train,
            n_test=test,
            x_std=0,
            t_std=0,
        )

    def _make_split(self, n: int) -> DataSplit:
        return DataSplit(
            x=torch.rand([n, *self._input_shape]),
            t=torch.randint(
                low=0,
                high=self._output_shape[-1],
                size=[n, *self._output_shape[:-1]],
            ),
            n=n,
        )

    @classmethod
    def get_default_loss(cls) -> type[Loss] | None:
        return CrossEntropy

    @classmethod
    def get_cli_arg(cls):
        return cli.ObjectArg(
            cls,
            cli.Arg("input_shape",  type=int, nargs="+", default=[3, 32, 32]),
            cli.Arg("output_shape", type=int, nargs="+", default=[10]),
            cli.Arg("train",        type=int, default=200),
            cli.Arg("test",         type=int, default=200),
        )

class RandomRegression(RandomClassification):
    def _make_split(self, n: int) -> DataSplit:
        return DataSplit(
            x=torch.rand([n, *self._input_shape]),
            t=torch.rand([n, *self._output_shape]),
            n=n,
        )

    @classmethod
    def get_default_loss(cls) -> type[Loss] | None:
        return Mse



import os
import multiprocessing
import queue
from jutility import plotting, util, cli
from juml.train.base import Trainer
from juml.tools.experiment import ExperimentGroup

class Sweeper:
    def __init__(
        self,
        args:       cli.ParsedArgs,
        params:     dict[str, list],
        devices:    list[list[int]],
        seeds:      list[int],
        no_cache:   bool,
        log_x:      list[str],
        configs:    list[str],
        **train_kwargs,
    ):
        printer = util.Printer()
        printer.heading("Sweeper: Initialise experiments")

        Trainer.apply_configs(args, configs, list(params.keys()))
        self.params         = params
        self.seeds          = seeds
        self.log_x          = log_x
        self.experiments    = ExperimentGroup.from_params(params, seeds)
        self.init_metric_info(args)
        self.init_name(args)
        self.init_output_dir()

        printer.heading("Sweeper: Run experiments")

        if len(devices) == 1:
            [d] = devices
            cli.verbose.reset()
            self.run_single_process(
                args=args,
                devices=d,
                no_cache=no_cache,
                train_kwargs=train_kwargs,
                printer=printer,
            )
        else:
            self.run_multi_process(
                args=args,
                devices=devices,
                no_cache=no_cache,
                train_kwargs=train_kwargs,
            )

        printer.heading("Sweeper: display results")

        self.experiments.load_results(args, self.opt_str)
        self.best = (
            max(self.experiments)
            if self.maximise else
            min(self.experiments)
        )

        self.best_model_dir = self.best.metrics["model_dir"]
        self.plot_paths = [os.path.join(self.best_model_dir, "metrics.png")]

        for param_name in self.params.keys():
            self.plot_param(param_name)

        self.save_results_markdown()
        self.experiments.save(self.output_dir)

    def init_metric_info(self, args: cli.ParsedArgs):
        dataset = Trainer.init_dataset(args)
        loss    = Trainer.init_loss(args, dataset)
        self.metric_info    = loss.metric_info()
        self.maximise       = loss.metric_higher_is_better()
        self.opt_str        = "max" if self.maximise else "min"

    def init_name(self, args: cli.ParsedArgs):
        original_args = {k: args.get_value(k) for k in self.params.keys()}
        original_args["seed"] = args.get_value("seed")

        for i, e in enumerate(self.experiments, start=1):
            print("(%2i) %s" % (i, e.arg_str))
            args.update(e.arg_dict)
            e.set_model_name(Trainer.get_summary(args))
            e.set_ind(i)

        args.update(original_args)
        sorted_names = sorted(e.model_name for e in self.experiments)
        self.name = util.merge_strings(sorted_names)

    def init_output_dir(self):
        self.output_dir = os.path.join("results", "sweep", self.name)

    def run_single_process(
        self,
        args:           cli.ParsedArgs,
        devices:        list[int],
        no_cache:       bool,
        train_kwargs:   dict,
        printer:        util.Printer,
    ):
        for e in self.experiments:
            run_experiment(
                args=args,
                devices=devices,
                no_cache=no_cache,
                train_kwargs=train_kwargs,
                args_update_dict=e.arg_dict,
                printer=printer,
            )

    def run_multi_process(
        self,
        args:           cli.ParsedArgs,
        devices:        list[list[int]],
        no_cache:       bool,
        train_kwargs:   dict,
    ):
        mp_context = multiprocessing.get_context("spawn")

        q = mp_context.Queue()
        for e in self.experiments:
            q.put(e.arg_dict)

        p_list = [
            mp_context.Process(
                target=sweeper_subprocess,
                kwargs={
                    "args":         args,
                    "q":            q,
                    "pid":          i,
                    "devices":      d,
                    "no_cache":     no_cache,
                    "train_kwargs": train_kwargs,
                    "output_dir":   self.output_dir,
                },
            )
            for i, d in enumerate(devices)
        ]

        for p in p_list:
            p.start()

        for p in p_list:
            p.join()

    def plot_param(self, param_name: str):
        sweep = self.experiments.sweep_param(self.best, param_name)

        x_index = not sweep.is_numeric()
        log_x = (True if (param_name in self.log_x) else False)
        log_y = self.metric_info.get("log_y", False)
        results_train   = plotting.NoisyData(log_y=log_y, x_index=x_index)
        results_test    = plotting.NoisyData(log_y=log_y, x_index=x_index)
        results_time    = plotting.NoisyData(log_y=True,  x_index=x_index)
        results_size    = plotting.NoisyData(log_y=True,  x_index=x_index)

        for e in sweep:
            val = e.arg_dict[param_name]
            results_train.update(val,   e.metrics["train"][self.opt_str])
            results_test.update(val,    e.metrics["test" ][self.opt_str])
            results_time.update(val,    e.metrics["time" ])
            results_size.update(val,    e.metrics["num_params"])

        mp = plotting.MultiPlot(
            plotting.Subplot(
                results_train.plot(),
                results_train.plot_best(
                    maximise=self.maximise,
                    label_fmt="(%s, %.5f)",
                ),
                plotting.Legend(),
                **results_train.get_xtick_kwargs(),
                **self.metric_info,
                xlabel=param_name,
                log_x=log_x,
                title="Best train metric",
            ),
            plotting.Subplot(
                results_test.plot(),
                results_test.plot_best(
                    maximise=self.maximise,
                    label_fmt="(%s, %.5f)",
                ),
                plotting.Legend(),
                **results_test.get_xtick_kwargs(),
                **self.metric_info,
                xlabel=param_name,
                log_x=log_x,
                title="Best test metric",
            ),
            plotting.Subplot(
                results_time.plot(),
                **results_time.get_xtick_kwargs(),
                xlabel=param_name,
                log_x=log_x,
                log_y=True,
                ylabel="Time (s)",
                title="Training duration",
            ),
            plotting.Subplot(
                results_size.plot(),
                **results_size.get_xtick_kwargs(),
                xlabel=param_name,
                log_x=log_x,
                log_y=True,
                ylabel="Number of parameters",
                title="Model size",
            ),
            title="%s\n%r" % (param_name, self.params[param_name]),
            title_font_size=15,
            figsize=[10, 8],
        )
        full_path = mp.save(param_name, self.output_dir)
        self.plot_paths.append(full_path)

    def save_results_markdown(self):
        md = util.MarkdownPrinter("results", self.output_dir)
        md.title("Sweep results")

        md.heading("Summary", end="\n\n")
        table = util.Table.key_value(printer=md)
        table.update(k="`#` experiments",   v=len(self.experiments))
        table.update(k="Best result",       v="%.5f" % self.best.result)
        table.update(k="Target metric",     v="`test.%s`" % self.opt_str)
        for param_name, param_vals in self.params.items():
            table.update(k=md.code(param_name), v=md.code(str(param_vals)))

        table.update(k="`seeds`", v=md.code(str(self.seeds)))

        md.heading("Best model")
        best_rel_dir = os.path.relpath(self.best_model_dir, self.output_dir)
        best_args_json      = os.path.join(best_rel_dir, "args.json")
        best_metrics_json   = os.path.join(best_rel_dir, "metrics.json")
        md.file_link(best_args_json,    "Best args (JSON)")
        md.file_link(best_metrics_json, "Best metrics (JSON)", end="\n\n")
        table = util.Table.key_value(printer=md)
        table.update(k="Params/seed",  v=md.code(self.best.arg_str))

        for name, metric in [
            ("Model",               "repr_model"),
            ("Model name",          "model_name"),
            ("Train metrics",       "train_summary"),
            ("Test metrics",        "test_summary"),
            ("Training duration",   "time_str"),
            ("`#` parameters",      "num_params"),
        ]:
            table.update(k=name, v=md.code(self.best.metrics[metric]))

        for name, val in self.best.arg_dict.items():
            table.update(k="`--%s`" % name, v=md.code(val))

        md.heading("Statistics", end="\n\n")
        table = util.Table.key_value(md)
        best_config_sweep = self.experiments.sweep_seeds(self.best)
        mean_best = best_config_sweep.results_mean()
        mean_all = self.experiments.results_mean()
        table.update(k="Mean (best params)",    v="%.5f" % mean_best)
        table.update(k="Mean (all)",            v="%.5f" % mean_all)

        if len(self.seeds) >= 2:
            std_best = best_config_sweep.results_std()
            std_all = self.experiments.results_std()
            table.update(k="STD (best params)", v="%.5f" % std_best)
            table.update(k="STD (all)",         v="%.5f" % std_all)

        md.heading("Metrics")
        for full_path in self.plot_paths:
            md.image(os.path.relpath(full_path, self.output_dir))

        md.heading("All results", end="\n\n")
        table = util.Table(
            util.Column("rank",     "i",    width=-10),
            util.Column("result",   ".5f",  title="`test.%s`" % self.opt_str),
            *[
                util.Column(param_name, title=md.code(param_name))
                for param_name in self.params.keys()
            ],
            util.Column("seed", "i", title="`seed`"),
            util.Column("model_name"),
            printer=md,
        )
        sorted_experiments = sorted(self.experiments, reverse=self.maximise)
        for i, e in enumerate(sorted_experiments, start=1):
            metrics_png = os.path.join(e.metrics["model_dir"], "metrics.png")
            metrics_rel = os.path.relpath(metrics_png, self.output_dir)
            table.update(
                rank=i,
                result=e.result,
                model_name="[`%s`](%s)" % (e.model_name, metrics_rel),
                **e.arg_dict,
            )

        md.git_add(
            md.get_filename(),
            os.path.join(self.best_model_dir, "args.json"),
            os.path.join(self.best_model_dir, "metrics.json"),
            *self.plot_paths,
        )
        md.readme_include("`[ full_sweep_results ]`", *self.plot_paths)
        md.show_command("sweep")
        md.flush()

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.JsonArg(
                "params",
                default=dict(),
                metavar=": dict[str, list] = \"{}\"",
                help=(
                    "EG '{\"trainer.BpSp.epochs\":[100,200,300],"
                    "\"trainer.BpSp.optimiser.Adam.lr\":"
                    "[1e-5,1e-4,1e-3,1e-2]}'"
                )
            ),
            cli.JsonArg(
                "devices",
                default=[[]],
                metavar=": list[list[int]] = \"[[]]\"",
                help="EG \"[[1,2,3],[3,4],[5]]\" or \"[[],[],[],[],[],[]]\""
            ),
            cli.Arg(
                "seeds",
                type=int,
                nargs="+",
                default=list(range(5)),
            ),
            cli.Arg("no_cache", action="store_true"),
            cli.Arg("log_x",    type=str, default=[], nargs="+"),
        ]

def sweeper_subprocess(
    args:           cli.ParsedArgs,
    q:              multiprocessing.Queue,
    pid:            int,
    devices:        list[int],
    no_cache:       bool,
    train_kwargs:   dict,
    output_dir:     str,
):
    printer = util.Printer(
        "p%i_log" % pid,
        dir_name=output_dir,
        print_to_console=False,
    )
    printer("Devices = %s" % devices)
    while True:
        try:
            args_update_dict = q.get(block=False)
        except queue.Empty:
            return

        run_experiment(
            args=args,
            devices=devices,
            no_cache=no_cache,
            train_kwargs=train_kwargs,
            args_update_dict=args_update_dict,
            printer=printer,
        )

def run_experiment(
    args:               cli.ParsedArgs,
    devices:            list[int],
    no_cache:           bool,
    train_kwargs:       dict,
    args_update_dict:   dict,
    printer:            util.Printer,
):
    args.update(args_update_dict)
    for key in args_update_dict:
        if key in train_kwargs:
            train_kwargs[key] = args_update_dict[key]

    metrics_path = Trainer.get_metrics_path(args)
    if (not os.path.isfile(metrics_path)) or no_cache:
        with util.Timer(
            name=str(args_update_dict),
            printer=printer,
            hline=True,
        ):
            Trainer.from_args(
                args,
                devices=devices,
                configs=[],
                printer=printer,
                **train_kwargs,
            )
    else:
        print("Found cached results `%s` -> skip" % metrics_path)



import os
import statistics
from jutility import util, cli
from juml.train.base import Trainer

class Experiment:
    def __init__(
        self,
        arg_dict:   dict,
        model_name: (str    | None),
        ind:        (int    | None),
        metrics:    (dict   | None),
        result:     (float  | None),
    ):
        self.arg_dict   = arg_dict
        self.arg_str    = util.format_dict(arg_dict)
        self.model_name = model_name
        self.ind        = ind
        self.metrics    = metrics
        self.result     = result

    @classmethod
    def from_arg_dict(cls, arg_dict: dict):
        return cls(
            arg_dict=arg_dict,
            model_name=None,
            ind=None,
            metrics=None,
            result=None,
        )

    def to_dict(self) -> dict:
        return {
            "arg_dict":     self.arg_dict,
            "model_name":   self.model_name,
            "ind":          self.ind,
            "metrics":      self.metrics,
            "result":       self.result,
        }

    def set_model_name(self, model_name: str):
        self.model_name = model_name

    def set_ind(self, ind: int):
        self.ind = ind

    def load_result(self, args: cli.ParsedArgs, key: str):
        if self.result is not None:
            raise ValueError(
                "%r already has result=%s"
                % (self, self.result)
            )

        args.update(self.arg_dict)
        self.metrics    = util.load_json(Trainer.get_metrics_path(args))
        self.result     = self.metrics["test"][key]

        if not isinstance(self.result, float):
            raise ValueError(
                "metrics['test'][%r] = %s has type `%s`, expected `float`"
                % (key, self.result, type(self.result).__name__)
            )

    def __repr__(self) -> str:
        return util.format_type(type(self), **self.arg_dict)

    def __lt__(self, other: "Experiment") -> bool:
        return (self.result < other.result)

class ExperimentGroup:
    def __init__(
        self,
        params:         dict[str, list],
        seeds:          list[int],
        experiments:    list[Experiment],
    ):
        self.params             = params
        self.seeds              = seeds
        self.experiment_list    = experiments
        self.experiment_dict    = {e.arg_str: e for e in experiments}

    @classmethod
    def from_params(
        cls,
        params: dict[str, list],
        seeds:  list[int],
    ):
        components_list = [[["seed", s]] for s in seeds]
        for param_name, param_vals in params.items():
            components_list = [
                c + p
                for c in components_list
                for p in [[[param_name, v]] for v in param_vals]
            ]

        return cls(
            params=params,
            seeds=seeds,
            experiments=[
                Experiment.from_arg_dict({k: v for k, v in c})
                for c in components_list
            ],
        )

    @classmethod
    def load(cls, sweep_name: str):
        dir_name    = os.path.join("results", "sweep", sweep_name)
        full_path   = util.get_full_path(
            "experiments.json",
            dir_name=dir_name,
            loading=True,
        )
        eg_dict = util.load_json(full_path)
        return cls(
            params=eg_dict["params"],
            seeds=eg_dict["seeds"],
            experiments=[
                Experiment(**ed)
                for ed in eg_dict["experiments"]
            ],
        )

    def save(self, output_dir: str):
        util.save_json(
            {
                "params":       self.params,
                "seeds":        self.seeds,
                "experiments":  [e.to_dict() for e in self.experiment_list],
            },
            "experiments",
            dir_name=output_dir,
        )

    def load_results(self, args: cli.ParsedArgs, key: str):
        for e in self.experiment_list:
            e.load_result(args, key)

    def sweep_seeds(
        self,
        root_experiment: Experiment,
    ) -> "ExperimentGroup":
        experiment_list = []
        root_dict       = root_experiment.arg_dict
        root_seed       = root_dict["seed"]

        for seed in self.seeds:
            root_dict["seed"] = seed
            arg_str = util.format_dict(root_dict)
            experiment_list.append(self.experiment_dict[arg_str])

        root_dict["seed"] = root_seed
        return ExperimentGroup(
            params=dict(),
            seeds=self.seeds,
            experiments=experiment_list,
        )

    def sweep_param(
        self,
        root_experiment:    Experiment,
        param_name:         str,
    ) -> "ExperimentGroup":
        experiment_list = []
        root_dict       = root_experiment.arg_dict
        root_val        = root_dict[param_name]
        root_seed       = root_dict["seed"]

        for val in self.params[param_name]:
            for seed in self.seeds:
                root_dict[param_name]   = val
                root_dict["seed"]       = seed
                arg_str = util.format_dict(root_dict)
                experiment_list.append(self.experiment_dict[arg_str])

        root_dict[param_name]   = root_val
        root_dict["seed"]       = root_seed
        return ExperimentGroup(
            params={param_name: self.params[param_name]},
            seeds=self.seeds,
            experiments=experiment_list,
        )

    def is_numeric(self) -> bool:
        return all(
            (isinstance(v, int) or isinstance(v, float))
            for param_vals in self.params.values()
            for v in param_vals
        )

    def get_results(self) -> list[float | None]:
        return [e.result for e in self.experiment_list]

    def results_mean(self) -> float:
        return statistics.mean(self.get_results())

    def results_std(self) -> float:
        return statistics.stdev(self.get_results())

    def __iter__(self):
        return iter(self.experiment_list)

    def __len__(self):
        return len(self.experiment_list)



import math
import torch
from jutility import cli
from juml.models.base import Model
from juml.models.linear import Linear

class Pooler(Model):
    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        raise NotImplementedError()

    def get_input_shape(self) -> list[int]:
        raise NotImplementedError()

    def get_input_dim(self, dim: int) -> int:
        input_shape = self.get_input_shape()
        return input_shape[dim]

    def unpool(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return []

class Identity(Pooler):
    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        self._output_shape = output_shape

    def get_input_shape(self) -> list[int]:
        return self._output_shape

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x

    def unpool(self, x: torch.Tensor) -> torch.Tensor:
        return x

class Unflatten(Pooler):
    def __init__(self, n: int):
        self._torch_module_init()
        self._n = n

    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        self._unflatten_shape = output_shape[-self._n:]
        self._input_shape = [
            *output_shape[:-self._n],
            math.prod(self._unflatten_shape),
        ]

    def get_input_shape(self) -> list[int]:
        return self._input_shape

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.unflatten(-1, self._unflatten_shape)

    def unpool(self, x: torch.Tensor) -> torch.Tensor:
        return x.flatten(-self._n, -1)

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [cli.Arg("n", type=int, default=None)]

class Average2d(Pooler):
    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        self.linear = torch.nn.Linear(input_shape[-3], output_shape[-1])

    def forward(self, x_nchw: torch.Tensor) -> torch.Tensor:
        x_nci = x_nchw.flatten(-2, -1)
        x_nc  = x_nci.mean(dim=-1)
        x_no  = self.linear.forward(x_nc)
        return x_no

class Max2d(Pooler):
    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        self.linear = torch.nn.Linear(input_shape[-3], output_shape[-1])

    def forward(self, x_nchw: torch.Tensor) -> torch.Tensor:
        x_nci = x_nchw.flatten(-2, -1)
        x_nc  = x_nci.max(dim=-1).values
        x_no  = self.linear.forward(x_nc)
        return x_no

class SoftmaxAverage2d(Pooler):
    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        self.f_p = Linear(input_shape[-3], 1)
        self.f_x = Linear(input_shape[-3], output_shape[-1])

    def forward(self, x_nchw: torch.Tensor) -> torch.Tensor:
        x_nic   = x_nchw.flatten(-2, -1).transpose(-1, -2)
        p_ni1   = self.f_p.forward(x_nic)
        p_n1i   = p_ni1.squeeze(-1).unsqueeze(-2)
        x_n1c   = torch.softmax(p_n1i, dim=-1) @ x_nic
        x_nc    = x_n1c.squeeze(-2)
        x_no    = self.f_x.forward(x_nc)
        return x_no

class LinearSet2d(Pooler):
    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        self.linear = Linear(input_shape[-3], output_shape[-1])

    def forward(self, x_nchw: torch.Tensor) -> torch.Tensor:
        x_npc = x_nchw.flatten(-2, -1).transpose(-1, -2)
        x_npo = self.linear.forward(x_npc)
        return x_npo

class GatedLinearSet2d(Pooler):
    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        self.f_p = Linear(input_shape[-3], 1)
        self.f_x = Linear(input_shape[-3], output_shape[-1])

    def forward(self, x_nchw: torch.Tensor) -> torch.Tensor:
        x_npc = x_nchw.flatten(-2, -1).transpose(-1, -2)
        p_np1 = self.f_p.forward(x_npc)
        x_npo = self.f_x.forward(x_npc)
        x_npo = torch.sigmoid(p_np1) * x_npo
        return x_npo

class SetAverage(Pooler):
    def set_shapes(
        self,
        input_shape:  list[int],
        output_shape: list[int],
    ):
        self.input_shape = input_shape
        self.linear = torch.nn.Linear(input_shape[-1], output_shape[-1])

    def get_input_shape(self) -> list[int]:
        return self.input_shape

    def forward(self, x_npd: torch.Tensor) -> torch.Tensor:
        x_nd = x_npd.mean(dim=-2)
        x_no = self.linear.forward(x_nd)
        return x_no

def get_types() -> list[type[Pooler]]:
    return [
        Identity,
        Unflatten,
        Average2d,
        Max2d,
        SoftmaxAverage2d,
        LinearSet2d,
        GatedLinearSet2d,
        SetAverage,
    ]

def get_cli_choice() -> cli.ObjectChoice:
    return cli.ObjectChoice(
        "pooler",
        *[pool_type.get_cli_arg() for pool_type in get_types()],
        default="Identity",
    )



import torch
from juml.loss.mse import Mse

class AlignedSetMse(Mse):
    def forward(self, y_npo: torch.Tensor, t_npo: torch.Tensor):
        e_npo = self.weights * (y_npo - t_npo)
        return e_npo.square().sum(dim=-1).mean()

    def metric_batch(self, y_npo: torch.Tensor, t_npo: torch.Tensor):
        e_npo = self.weights * (y_npo - t_npo)
        return e_npo.square().mean(dim=-2).sum().item()

    @classmethod
    def needs_weights(cls) -> bool:
        return True



import os
from jutility import cli, util
from juml.commands.base import Command
from juml.device import DeviceConfig
from juml.train.base import Trainer
from juml.tools.display import plot_sequential

class PlotSequential(Command):
    @classmethod
    def run(
        cls,
        args:           cli.ParsedArgs,
        batch_size:     int,
        num_warmup:     int,
        devices:        list[int],
    ):
        device_cfg = DeviceConfig(devices)
        model_dir, model, dataset = Trainer.load(args)
        device_cfg.set_module_device(model)

        train_loader = dataset.get_data_loader("train", batch_size)
        x, t = next(iter(train_loader))
        [x] = device_cfg.to_device([x])

        for _ in range(num_warmup):
            y = model.forward(x)

        name = cls.get_name()
        md = util.MarkdownPrinter(name, model_dir)
        md.title(md.code(repr(model)), end="\n\n")
        md.set_print_to_console(True)

        mp = plot_sequential(model, x, md)
        mp.save(name, model_dir)

        md.set_print_to_console(False)
        md.image(name + ".png")

        md.git_add(md.get_filename(), mp.full_path)
        md.readme_include("`[ %s ]`" % repr(model))
        md.show_command("plotsequential")
        md.flush()

        return mp

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("batch_size",   type=int, default=100),
            cli.Arg("num_warmup",   type=int, default=10),
            cli.Arg("devices",      type=int, default=[], nargs="*"),
        ]



from jutility import plotting, cli

class PlotType:
    @classmethod
    def plot(cls, mp: plotting.MultiPlot, name: str, output_dir: str):
        raise NotImplementedError()

    @classmethod
    def get_cli_arg(cls):
        return cli.ObjectChoice(
            "plot_type",
            cli.ObjectArg(Png),
            cli.ObjectArg(Pdf),
            cli.ObjectArg(Show),
            default="Png",
            is_kwarg=True,
        )

class Png(PlotType):
    @classmethod
    def plot(cls, mp: plotting.MultiPlot, name: str, output_dir: str):
        mp.save(name, output_dir)

class Pdf(PlotType):
    @classmethod
    def plot(cls, mp: plotting.MultiPlot, name: str, output_dir: str):
        plotting.set_latex_params(use_times=True)
        mp.save(name, output_dir, pdf=True)

class Show(PlotType):
    @classmethod
    def plot(cls, mp: plotting.MultiPlot, name: str, output_dir: str):
        mp.save(name, output_dir, close=False)
        mp.show()



import torch
from juml.loss.base import Loss

class CrossEntropy(Loss):
    def forward(self, y, t):
        return torch.nn.functional.cross_entropy(y, t)

    def metric_batch(self, y, t):
        return torch.where(y.argmax(dim=-1) == t, 1, 0).sum().item()

    @classmethod
    def info(cls):
        return {"ylabel": "Loss"}

    @classmethod
    def metric_info(cls):
        return {"ylabel": "Accuracy", "ylim": [0, 1]}

    @classmethod
    def metric_higher_is_better(cls) -> bool:
        return True



import torch
from jutility import cli
from juml.loss.base import Loss

class ChamferMse(Loss):
    def chamfer_components(self, y_nqc: torch.Tensor, t_npc: torch.Tensor):
        t_np1c  = t_npc.unsqueeze(-2)
        y_n1qc  = y_nqc.unsqueeze(-3)
        e_npqc  = self.weights * (t_np1c - y_n1qc)
        mse_npq = e_npqc.square().sum(dim=-1)
        mse_np  = mse_npq.min(dim=-1).values
        mse_nq  = mse_npq.min(dim=-2).values
        return mse_np, mse_nq

    def forward(self, y, t):
        mse_np, mse_nq = self.chamfer_components(y, t)
        return mse_np.mean() + mse_nq.mean()

    def metric_batch(self, y, t):
        mse_np, mse_nq = self.chamfer_components(y, t)
        return (
            mse_np.mean(dim=-1).sum().item() +
            mse_nq.mean(dim=-1).sum().item()
        )

    @classmethod
    def info(cls):
        return {"ylabel": "Chamfer MSE", "log_y": True}

    @classmethod
    def needs_weights(cls) -> bool:
        return True

    @classmethod
    def get_cli_arg(cls):
        return cli.ObjectArg(cls, tag="CH")



import torch
from jutility import util, cli, units

class Model(torch.nn.Module):
    def _torch_module_init(self):
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def init_batch(
        self,
        x: torch.Tensor,
        t: (torch.Tensor | None),
        **kwargs,
    ):
        return

    def num_params(self) -> int:
        return sum(
            int(p.numel())
            for p in self.parameters()
            if  p.requires_grad
        )

    @classmethod
    def get_cli_arg(cls):
        return cli.ObjectArg(
            cls,
            *cls.get_cli_options(),
            tag=cls.get_tag(),
        )

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        raise NotImplementedError()

    @classmethod
    def get_tag(cls) -> (str | None):
        return None

    def __repr__(self):
        return util.format_type(
            type(self),
            num_params=units.metric.format(self.num_params()),
            item_fmt="%s=%s",
        )



import torch
import torch.utils.data
from jutility import util, cli
from juml.device import DeviceConfig

class Loss:
    def __init__(self):
        self.weights = None

    def forward(self, y: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def metric_batch(self, y: torch.Tensor, t: torch.Tensor) -> float:
        raise NotImplementedError()

    @classmethod
    def info(cls) -> dict:
        raise NotImplementedError()

    @classmethod
    def metric_info(cls) -> dict:
        return cls.info()

    @classmethod
    def metric_higher_is_better(cls) -> bool:
        return False

    @classmethod
    def get_optimum_metric(cls, *args):
        return max(*args) if cls.metric_higher_is_better() else min(*args)

    @classmethod
    def get_optimum_metric_str(cls) -> str:
        return "max" if cls.metric_higher_is_better() else "min"

    def metric(
        self,
        model:          torch.nn.Module,
        data_loader:    torch.utils.data.DataLoader,
        device_cfg:     DeviceConfig,
    ) -> float:
        metric_sum = 0
        for x, t in data_loader:
            x, t = device_cfg.to_device([x, t])
            y = model.forward(x)
            metric_sum += self.metric_batch(y, t)

        return metric_sum / len(data_loader.dataset)

    def needs_weights(self) -> bool:
        return False

    def set_weights(self, weights: torch.Tensor):
        self.weights = weights

    def set_device(self, device_cfg: DeviceConfig):
        if self.weights is not None:
            [self.weights] = device_cfg.to_device([self.weights])

    @classmethod
    def get_cli_arg(cls):
        return cli.ObjectArg(cls)

    def __repr__(self) -> str:
        return util.format_type(type(self), weights=self.weights)



from juml.models.base import Model
from juml.datasets.base import Dataset
from juml.loss.base import Loss
from juml.train.base import Trainer
from juml.commands.base import Command
from juml.framework import Framework



import torch
import torch.utils.data
from jutility import cli
from juml.datasets.split import DataSplit
from juml.datasets.fromdict import DatasetFromDict

class Synthetic(DatasetFromDict):
    def _init_synthetic(
        self,
        input_shape:    list[int],
        output_shape:   list[int],
        n_train:        int,
        n_test:         int,
        x_std:          float,
        t_std:          float,
    ):
        self._input_shape   = input_shape
        self._output_shape  = output_shape
        self._n_train       = n_train
        self._n_test        = n_test
        self._x_std         = x_std
        self._t_std         = t_std
        self._split_dict    = {
            "train": self._make_split(self._n_train),
            "test":  self._make_split(self._n_test),
        }

    def _make_split(self, n: int) -> DataSplit:
        x = self._sample_input(n)
        t = self._compute_target(x)
        return DataSplit(
            x=x + torch.normal(0, self._x_std, x.shape),
            t=t + torch.normal(0, self._t_std, t.shape),
            n=n,
        )

    def _sample_input(self, n: int) -> torch.Tensor:
        return torch.normal(0, 1, [n, *self._input_shape])

    def _compute_target(self, x: torch.Tensor) -> torch.Tensor:
        raise NotImplementedError()

    def get_loss_weights(self) -> torch.Tensor:
        return 1.0 / self._split_dict["train"].t.flatten(0, -2).std(dim=-2)

    def get_input_shape(self) -> list[int]:
        return self._input_shape

    def get_output_shape(self) -> list[int]:
        return self._output_shape



import torch
from jutility import plotting, util, units
from juml.models.sequential import Sequential

def display_sequential(
    model:      Sequential,
    x:          torch.Tensor,
    printer:    (util.Printer | None)=None,
) -> tuple[torch.Tensor, util.Table]:
    util.hline()
    w = max(len(repr(m)) for m in model.layers)
    table = util.Table(
        util.Column("layer",    "r",    -w,     "Layer"),
        util.Column("shape",    "s",    -22,    "Output shape"),
        util.Column("t",        ".5fs", 10,     "Time"),
        printer=printer,
    )
    total_timer = util.Timer(verbose=False)
    layer_timer = util.Timer(verbose=False)
    display_layer(table, "Input", x, layer_timer)

    with layer_timer:
        x = model.embed.forward(x)
        display_layer(table, model.embed, x, layer_timer)
    for layer in model.layers:
        with layer_timer:
            x = layer.forward(x)
            display_layer(table, layer, x, layer_timer)
    with layer_timer:
        x = model.pool.forward(x)
        display_layer(table, model.pool, x, layer_timer)

    table.update(layer="Full model:")
    display_layer(table, model, x, total_timer)
    util.hline()
    return x, table

def display_layer(
    table:  util.Table,
    layer:  (torch.nn.Module | str),
    x:      torch.Tensor,
    timer:  util.Timer,
):
    table.update(layer=layer, shape=list(x.shape), t=timer.get_time_taken())

def num_params(layer: torch.nn.Module) -> int:
    return sum(
        int(p.numel())
        for p in layer.parameters()
        if  p.requires_grad
    )

def plot_sequential(
    model:      Sequential,
    x:          torch.Tensor,
    printer:    (util.Printer | None)=None,
) -> plotting.MultiPlot:
    _, table    = display_sequential(model, x, printer)
    t_data      = table.get_data("t")
    t_list      = t_data[1:]
    t_tot       = t_list[-1]
    t_max       = max(t_list[:-1])
    t_tot_label = "Total = %s"  % units.time_concise.format(t_tot)
    t_max_label = "Max = %s"    % units.time_concise.format(t_max)

    layer_list  = [model.embed, *model.layers, model.pool]
    np_list     = [num_params(m) for m in layer_list]
    n_tot_label = "Total = %s"  % units.metric.format(sum(np_list))
    n_max_label = "Max = %s"    % units.metric.format(max(np_list))

    name_list   = [type(m).__name__ for m in layer_list]
    x_plot      = list(range(len(name_list)))
    kwargs      = {
        "xticks":               x_plot,
        "xticklabels":          name_list,
        "rotate_xticklabels":   True,
        "xlabel":               "Layer",
    }
    return plotting.MultiPlot(
        plotting.Subplot(
            plotting.Bar(x_plot, t_list[:-1]),
            plotting.HLine(t_tot, c="k", ls="--", label=t_tot_label),
            plotting.HLine(t_max, c="r", ls="--", label=t_max_label),
            plotting.Legend(),
            **kwargs,
            ylim=[0, 1.1 * t_max],
            ylabel="Time (s)",
        ),
        plotting.Subplot(
            plotting.Bar(x_plot, np_list),
            plotting.HLine(sum(np_list), c="k", ls="--", label=n_tot_label),
            plotting.HLine(max(np_list), c="r", ls="--", label=n_max_label),
            plotting.Legend(),
            **kwargs,
            ylim=[0, 1.1 * max(np_list)],
            ylabel="# Parameters",
        ),
        title="%r\nInput shape = %s" % (model, list(x.shape)),
        figsize=[10, 6],
    )



import torch
from juml.loss.base import Loss

class Mse(Loss):
    def forward(self, y, t):
        return (y - t).square().sum(dim=-1).mean()

    def metric_batch(self, y, t):
        return (y - t).square().sum().item()

    @classmethod
    def info(cls):
        return {"ylabel": "MSE", "log_y": True}



import math
import torch
from juml.models.base import Model

class Linear(Model):
    def __init__(
        self,
        input_dim:  int,
        output_dim: int,
        w_scale:    (float | None)=None,
    ):
        if w_scale is None:
            w_scale = 1 / math.sqrt(input_dim)

        w_shape = [input_dim, output_dim]
        self._torch_module_init()
        self.w_io = torch.nn.Parameter(torch.normal(0, w_scale, w_shape))
        self.b_o  = torch.nn.Parameter(torch.zeros(output_dim))

    def forward(self, x_ni: torch.Tensor) -> torch.Tensor:
        x_no = x_ni @ self.w_io + self.b_o
        return x_no

    def init_batch(
        self,
        x:      torch.Tensor,
        t:      (torch.Tensor | None),
        eps:    float=1e-5,
    ):
        with torch.no_grad():
            self.b_o.zero_()
            if t is None:
                x = x.flatten(0, -2)
                self.w_io *= 1 / (eps + x.std(-2).unsqueeze(-1))
                y = self.forward(x)
                self.w_io *= 1 / (eps + y.std(-2).unsqueeze(-2))
                y = self.forward(x)
                y_lo = y.quantile(0.01, dim=-2)
                y_hi = y.quantile(0.99, dim=-2)
                r = torch.rand(self.b_o.shape)
                t = y_lo + r * (y_hi - y_lo)
                self.b_o.copy_(-t)
            else:
                x = x.flatten(0, -2)
                t = t.flatten(0, -2)
                xm = x.mean(-2)
                tm = t.mean(-2)
                w, _, _, _ = torch.linalg.lstsq(x - xm, t - tm)
                self.w_io.copy_(w)
                self.b_o.copy_(tm - self.forward(xm))



import torch.utils.data
from juml.datasets.base import Dataset

class DatasetFromDict(Dataset):
    def __init__(self):
        self._init_split_dict()

    def _init_split_dict(self):
        self._split_dict = self._get_split_dict()

    def _get_split_dict(self) -> dict[str, torch.utils.data.Dataset]:
        raise NotImplementedError()

    def get_data_split(self, split: str) -> torch.utils.data.Dataset:
        return self._split_dict[split]

    def get_split_names(self):
        s = sorted(self._split_dict.keys())
        s.insert(0, s.pop(s.index("train")))
        s.insert(1, s.pop(s.index("test")))
        return s



import os
import torch
from jutility import util

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))

def torch_set_print_options(
    precision:  int=3,
    threshold:  (int | float)=1e3,
    linewidth:  (int | float)=1e5,
    sci_mode:   bool=False,
):
    torch.set_printoptions(
        precision=precision,
        threshold=int(threshold),
        linewidth=int(linewidth),
        sci_mode=sci_mode,
    )

def get_output_dir(*subdir_names: str):
    return os.path.join("tests", "Outputs", *subdir_names)

def set_torch_seed(*args):
    seed = util.Seeder().get_seed(*args)
    torch.manual_seed(seed)

class TensorPrinter:
    def __init__(self, printer: (util.Printer | None)=None):
        if printer is None:
            printer = util.Printer()

        self.printer = printer

    def __call__(self, x: torch.Tensor, heading: (str | None)=None):
        if heading is not None:
            self.printer.heading(heading)

        self.printer("\n```\n%s\n%s\n%s\n```" % (x.numel(), x.shape, x))



import torch
from jutility import cli, plotting
from juml.commands.base import Command
from juml.train.base import Trainer

class Plot1dRegression(Command):
    @classmethod
    def run(
        cls,
        args:   cli.ParsedArgs,
        n_plot: int,
    ):
        model_dir, model, dataset = Trainer.load(args)

        data = dict()
        xlim = []
        for split in ["train", "test"]:
            data_split  = dataset.get_data_split(split)
            data_loader = dataset.get_data_loader(split, len(data_split))
            x, t = next(iter(data_loader))
            assert isinstance(x, torch.Tensor)
            assert isinstance(t, torch.Tensor)
            data[split] = [x.flatten(), t.flatten()]
            xlim.extend([x.min().item(), x.max().item()])

        x = torch.linspace(min(xlim), max(xlim), n_plot).reshape(n_plot, 1)
        y = model.forward(x).detach().reshape(n_plot)

        mp = plotting.MultiPlot(
            plotting.Subplot(
                plotting.Scatter(*data["train"], c="b", label="train"),
                plotting.Scatter(*data["test" ], c="r", label="test" ),
                plotting.Line(x, y, c="g", label="Prediction"),
                plotting.Legend(),
            ),
            figsize=[8, 6],
            title="%r\n%r" % (model, dataset),
            title_font_size=15,
        )
        mp.save("Predictions", model_dir)

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("n_plot", type=int, default=200),
        ]



import torch.utils.data
import torchvision
from juml.datasets import DATA_REL_DIR
from juml.datasets.fromdict import DatasetFromDict
from juml.loss.base import Loss
from juml.loss.crossentropy import CrossEntropy

class Cifar10(DatasetFromDict):
    def _get_split_dict(self) -> dict[str, torch.utils.data.Dataset]:
        return {
            "train": torchvision.datasets.CIFAR10(
                root=DATA_REL_DIR,
                train=True,
                transform=torchvision.transforms.ToTensor(),
                download=True,
            ),
            "test": torchvision.datasets.CIFAR10(
                root=DATA_REL_DIR,
                train=False,
                transform=torchvision.transforms.ToTensor(),
                download=True,
            ),
        }

    def get_input_shape(self) -> list[int]:
        return [3, 32, 32]

    def get_output_shape(self) -> list[int]:
        return [10]

    @classmethod
    def get_default_loss(cls) -> type[Loss] | None:
        return CrossEntropy



import os
import torch

class DeviceConfig:
    def __init__(self, devices: list[int]):
        self._has_devices = (len(devices) > 0)
        if self._has_devices:
            devices_str = ",".join(str(d) for d in devices)
            os.environ["CUDA_VISIBLE_DEVICES"] = devices_str

    def set_module_device(self, module: torch.nn.Module):
        if self._has_devices:
            module.cuda()

    def to_device(
        self,
        input_tensors: list[torch.Tensor],
    ) -> list[torch.Tensor]:
        if self._has_devices:
            input_tensors = [x.cuda() for x in input_tensors]

        return input_tensors



import torch
import torch.utils.data
from jutility import util

class DataSplit(torch.utils.data.Dataset):
    def __init__(self, x: torch.Tensor, t: torch.Tensor, n: int):
        self.x = x
        self.t = t
        self.n = n

    def __getitem__(self, index):
        return self.x[index], self.t[index]

    def __len__(self):
        return self.n

    def __repr__(self):
        return util.format_type(
            type(self),
            x_shape=list(self.x.shape),
            t_shape=list(self.t.shape),
            n=self.n,
        )



import torch
from jutility import cli
from juml.models.base import Model
from juml.models.sequential import Sequential
from juml.models.embed import Embedder
from juml.models.pool import Pooler

class Cnn(Sequential):
    def __init__(
        self,
        input_shape:        list[int],
        output_shape:       list[int],
        kernel_size:        int,
        channel_dim:        int,
        num_stages:         int,
        blocks_per_stage:   int,
        stride:             int,
        embedder:           Embedder,
        pooler:             Pooler,
    ):
        self._init_sequential(embedder, pooler)
        self.embed.set_input_shape(input_shape)
        self.pool.set_shapes([channel_dim, None, None], output_shape)

        input_dim = self.embed.get_output_dim(-3)
        layer = InputReluCnnLayer(input_dim, channel_dim, kernel_size, stride)
        self.layers.append(layer)

        for _ in range(num_stages - 1):
            for _ in range(blocks_per_stage - 1):
                layer = ReluCnnLayer(channel_dim, kernel_size)
                self.layers.append(layer)

            layer = StridedReluCnnLayer(channel_dim, kernel_size, stride)
            self.layers.append(layer)

        for _ in range(blocks_per_stage):
            layer = ReluCnnLayer(channel_dim, kernel_size)
            self.layers.append(layer)

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("kernel_size",      type=int, default=5),
            cli.Arg("channel_dim",      type=int, default=64),
            cli.Arg("num_stages",       type=int, default=3),
            cli.Arg("blocks_per_stage", type=int, default=2),
            cli.Arg("stride",           type=int, default=2),
        ]

class ReluCnnLayer(Model):
    def __init__(
        self,
        channel_dim: int,
        kernel_size: int,
    ):
        self._torch_module_init()
        self.conv = torch.nn.Conv2d(
            in_channels=channel_dim,
            out_channels=channel_dim,
            kernel_size=kernel_size,
            padding="same",
        )

    def forward(self, x_nchw: torch.Tensor) -> torch.Tensor:
        x_nchw = self.conv.forward(x_nchw)
        x_nchw = torch.relu(x_nchw)
        return x_nchw

class InputReluCnnLayer(ReluCnnLayer):
    def __init__(
        self,
        input_channel_dim:  int,
        output_channel_dim: int,
        kernel_size:        int,
        stride:             int,
    ):
        self._torch_module_init()
        self.conv = torch.nn.Conv2d(
            in_channels=input_channel_dim,
            out_channels=output_channel_dim,
            kernel_size=kernel_size,
            stride=stride,
        )

class StridedReluCnnLayer(ReluCnnLayer):
    def __init__(
        self,
        channel_dim:    int,
        kernel_size:    int,
        stride:         int,
    ):
        self._torch_module_init()
        self.conv = torch.nn.Conv2d(
            in_channels=channel_dim,
            out_channels=channel_dim,
            kernel_size=kernel_size,
            stride=stride,
        )



from jutility import cli
from juml.commands.base import Command
from juml.train.base import Trainer
from juml.tools.profiler import Profiler

class Profile(Command):
    @classmethod
    def run(
        cls,
        args:           cli.ParsedArgs,
        batch_size:     int,
        num_warmup:     int,
        num_profile:    int,
        devices:        list[int],
        sort_by:        str,
    ):
        model_dir, model, dataset = Trainer.load(args)
        return Profiler(
            model=model,
            dataset=dataset,
            model_dir=model_dir,
            batch_size=batch_size,
            num_warmup=num_warmup,
            num_profile=num_profile,
            devices=devices,
            name="profile",
            sort_by=sort_by,
        )

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("batch_size",   type=int, default=100),
            cli.Arg("num_warmup",   type=int, default=10),
            cli.Arg("num_profile",  type=int, default=10),
            cli.Arg("devices",      type=int, default=[], nargs="*"),
            cli.Arg("sort_by",      type=str, default="self_cpu_time_total"),
        ]



DATA_REL_DIR = "./data"

from juml.datasets.base import Dataset
from juml.datasets.split import DataSplit
from juml.datasets.fromdict import DatasetFromDict
from juml.datasets.synthetic import Synthetic
from juml.datasets.lineardataset import LinearDataset
from juml.datasets.sinmix import SinMix
from juml.datasets.randimg import RandomClassification, RandomRegression
from juml.datasets.mnist import Mnist
from juml.datasets.cifar10 import Cifar10

def get_all() -> list[type[Dataset]]:
    return [
        LinearDataset,
        SinMix,
        RandomClassification,
        RandomRegression,
        Mnist,
        Cifar10,
    ]




from jutility import cli
from juml.models.sequential import Sequential
from juml.models.embed import Embedder
from juml.models.pool import Pooler
from juml.models.linear import Linear

class LinearModel(Sequential):
    def __init__(
        self,
        input_shape:    list[int],
        output_shape:   list[int],
        embedder:       Embedder,
        pooler:         Pooler,
    ):
        self._init_sequential(embedder, pooler)
        self.embed.set_input_shape(input_shape)
        self.pool.set_shapes([], output_shape)

        layer = Linear(
            input_dim=self.embed.get_output_dim(-1),
            output_dim=self.pool.get_input_dim(-1),
        )
        self.layers.append(layer)

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return []



import os
from jutility import cli, util, plotting
from juml.commands.base import Command
from juml.datasets.base import Dataset
from juml.loss.base import Loss
from juml.tools.experiment import Experiment, ExperimentGroup
from juml.tools.plot_type import PlotType

class CompareSweeps(Command):
    @classmethod
    def run(
        cls,
        args:       cli.ParsedArgs,
        config:     list[dict[str, str]],
        xlabel:     str,
        clabel:     str,
        name:       (str | None),
        log_x:      bool,
        plot_type:  PlotType,
    ):
        dataset_type = args.get_type("dataset")
        assert issubclass(dataset_type, Dataset)

        loss_arg = args.get_arg("loss")
        loss_arg.set_default_choice(dataset_type.get_default_loss())
        loss_type = args.get_type("loss")
        assert issubclass(loss_type, Loss)

        metric_info = loss_type.metric_info()
        maximise    = loss_type.metric_higher_is_better()
        log_y       = metric_info.get("log_y", False)
        opt_str     = "max" if maximise else "min"
        cp          = plotting.ColourPicker(len(config), cyclic=True, cmap_name="gist_rainbow")

        series = [
            SweepSeries(
                **c,
                maximise=maximise,
                log_y=log_y,
                opt_str=opt_str,
                colour=cp.next(),
            )
            for c in config
        ]

        if name is None:
            name = util.merge_strings([s.sweep_name for s in series])

        output_dir = os.path.join("results", "compare", name)

        x_index = any((not s.experiments.is_numeric()) for s in series)
        xtick_config = plotting.NoisyData(x_index=x_index)
        for s in series:
            for param_vals in s.experiments.params.values():
                for v in param_vals:
                    if v != 30000:
                        xtick_config.update(v, None)

        axis_kwargs = {"xlabel": xlabel, "log_x": log_x}
        axis_kwargs.update(xtick_config.get_xtick_kwargs())
        metric_info["ylabel"] = "Test %s" % metric_info["ylabel"]

        mp = plotting.MultiPlot(
            # plotting.Subplot(
            #     *[
            #         s.plot(x_index, TrainGetter())
            #         for s in series
            #     ],
            #     **axis_kwargs,
            #     **metric_info,
            #     title="Best train metric",
            # ),
            plotting.Subplot(
                *[
                    s.plot(x_index, TestGetter())
                    for s in series
                ],
                **axis_kwargs,
                **metric_info,
            ),
            # plotting.Subplot(
            #     *[
            #         s.plot(x_index, TimeGetter())
            #         for s in series
            #     ],
            #     **axis_kwargs,
            #     log_y=True,
            #     ylabel="Time (s)",
            #     title="Training duration",
            # ),
            # plotting.Subplot(
            #     *[
            #         s.plot(x_index, SizeGetter())
            #         for s in series
            #     ],
            #     **axis_kwargs,
            #     log_y=True,
            #     ylabel="Number of parameters",
            #     title="Model size",
            # ),
            legend=plotting.FigureLegend(
                *[s.get_legend_plottable() for s in series],
                num_rows=None,
                loc="outside right upper",
                title=clabel,
            ),
            sharey=True,
            figsize=[6, 4],
        )
        plot_type.plot(mp, xlabel, output_dir)

        md = util.MarkdownPrinter(xlabel, output_dir)
        md.title("Sweep comparison")
        md.heading("Metrics")
        md.image(md.rel_path(mp.full_path))
        md.heading("Sweeps", end="\n\n")
        table = util.Table.key_value(printer=md)
        for s in series:
            table.update(
                k=s.series_name,
                v="[`[ %s ]`](../../sweep/%s/results.md)" % (
                    s.sweep_name,
                    s.sweep_name,
                ),
            )

        md.git_add(
            md.get_filename(),
            mp.full_path,
            *["results/sweep/%s/results.md" % s.sweep_name for s in series],
        )
        md.readme_include("`[ compare_sweeps ]`", mp.full_path)
        md.show_command("comparesweeps")
        md.flush()

        return mp

    @classmethod
    def include_arg(cls, arg: cli.Arg) -> bool:
        return arg.name in ["dataset", "loss"]

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.JsonArg(
                "config",
                required=True,
                metavar=": list[dict[\"series_name\": str, "
                "\"sweep_name\": str, \"param_name\": str]]",
                help=(
                    "EG '[{\"series_name\":\"MLP\","
                    "\"sweep_name\":\"mlp_sweep_name_s1,2,3\","
                    "\"param_name\":\"model.RzMlp.depth\"},"
                    "{\"series_name\":\"CNN\","
                    "\"sweep_name\":\"cnn_sweep_name_s1,2,3\","
                    "\"param_name\":\"model.RzCnn.depth\"}]'"
                )
            ),
            cli.Arg("xlabel",   type=str, required=True),
            cli.Arg("clabel",   type=str, required=True),
            cli.Arg("name",     type=str, default=None),
            cli.Arg("log_x",    action="store_true"),
            PlotType.get_cli_arg(),
        ]

class SweepSeries:
    def __init__(
        self,
        series_name:    str,
        sweep_name:     str,
        param_name:     str,
        maximise:       bool,
        log_y:          bool,
        opt_str:        str,
        colour:         list[float],
    ):
        self.series_name    = series_name
        self.sweep_name     = sweep_name
        self.param_name     = param_name
        self.log_y          = log_y
        self.opt_str        = opt_str
        self.colour         = colour

        eg_all              = ExperimentGroup.load(sweep_name)
        best                = max(eg_all) if maximise else min(eg_all)
        self.experiments    = eg_all.sweep_param(best, param_name)

    def plot(self, x_index: bool, mg: "MetricGetter") -> plotting.Plottable:
        nd = plotting.NoisyData(log_y=self.log_y, x_index=x_index)
        for e in self.experiments:
            x = e.arg_dict[self.param_name]
            y = mg.get(e.metrics, self.opt_str)
            nd.update(x, y)

        return nd.plot(c=self.colour)

    def get_legend_plottable(self) -> plotting.Plottable:
        nd = plotting.NoisyData()
        return nd.plot(c=self.colour, label=self.series_name)

class MetricGetter:
    def get(self, metrics: dict, opt_str: str) -> float:
        raise NotImplementedError()

class TrainGetter(MetricGetter):
    def get(self, metrics: dict, opt_str: str) -> float:
        return metrics["train"][opt_str]

class TestGetter(MetricGetter):
    def get(self, metrics: dict, opt_str: str) -> float:
        return metrics["test"][opt_str]

class TimeGetter(MetricGetter):
    def get(self, metrics: dict, opt_str: str) -> float:
        return metrics["time"]

class SizeGetter(MetricGetter):
    def get(self, metrics: dict, opt_str: str) -> float:
        return metrics["num_params"]



from juml.loss.base import Loss
from juml.loss.crossentropy import CrossEntropy
from juml.loss.mse import Mse
from juml.loss.asmse import AlignedSetMse
from juml.loss.chamfer import ChamferMse

def get_all() -> list[type[Loss]]:
    return [
        CrossEntropy,
        Mse,
        AlignedSetMse,
        ChamferMse,
    ]



import math
import torch
from jutility import cli
from juml.models.base import Model

class Embedder(Model):
    def set_input_shape(self, input_shape: list[int]):
        raise NotImplementedError()

    def get_output_shape(self) -> list[int]:
        raise NotImplementedError()

    def get_output_dim(self, dim: int) -> int:
        output_shape = self.get_output_shape()
        return output_shape[dim]

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return []

class Identity(Embedder):
    def set_input_shape(self, input_shape: list[int]):
        self._input_shape = input_shape

    def get_output_shape(self) -> list[int]:
        return self._input_shape

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x

class Flatten(Embedder):
    def __init__(self, n: int):
        self._torch_module_init()
        self._n = n

    def set_input_shape(self, input_shape: list[int]):
        self._input_shape = input_shape

    def get_output_shape(self) -> list[int]:
        return [
            *self._input_shape[:-self._n],
            math.prod(self._input_shape[-self._n:]),
        ]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.flatten(-self._n, -1)

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [cli.Arg("n", type=int, default=None)]

class CoordConv(Embedder):
    def set_input_shape(self, input_shape: list[int]):
        self._torch_module_init()
        c, h, w = input_shape[-3:]
        x_1w    = torch.linspace(-1, 1, w).unsqueeze(-2)
        y_h1    = torch.linspace(-1, 1, h).unsqueeze(-1)
        x_hw    = torch.tile(x_1w, [h, 1])
        y_hw    = torch.tile(y_h1, [1, w])
        xy_2hw  = torch.stack([x_hw, y_hw], dim=-3)
        self._coord_tensor = torch.nn.Parameter(xy_2hw, requires_grad=False)
        self._coord_shape  = list(self._coord_tensor.shape)
        self._output_shape = input_shape[:-3] + [c + 2, h, w]

    def get_output_shape(self) -> list[int]:
        return self._output_shape

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        batched_shape = list(x.shape[:-3]) + self._coord_shape
        batched_coord_tensor = self._coord_tensor.expand(batched_shape)
        return torch.concat([x, batched_coord_tensor], dim=-3)

def get_types() -> list[type[Embedder]]:
    return [
        Identity,
        Flatten,
        CoordConv,
    ]

def get_cli_choice() -> cli.ObjectChoice:
    return cli.ObjectChoice(
        "embedder",
        *[embed_type.get_cli_arg() for embed_type in get_types()],
        default="Identity",
    )



from jutility import cli, util
from juml.commands.base import Command
from juml.commands.compare_sweeps import CompareSweeps
from juml.tools.sweeper import Sweeper

class Sweep2d(Command):
    @classmethod
    def run(
        cls,
        args:   cli.ParsedArgs,
        name_1: str,
        name_2: str,
        vals_1: list,
        vals_2: list,
        log_x:  list[str],
        **kwargs,
    ):
        kwargs["log_x"] = log_x
        printer = util.Printer()
        printer.heading("Sweep2d: Sweep all")
        Sweeper(
            args=args,
            **kwargs,
            **args.get_kwargs(),
            params={
                name_1: vals_1,
                name_2: vals_2,
            },
        )
        kwargs["no_cache"] = False
        printer.heading("Sweep2d: Sweep %s" % name_1)
        sweep_configs_1 = cls.sweep_1d_loop(
            args=args,
            name_sweep=name_1,
            vals_sweep=vals_1,
            name_loop=name_2,
            vals_loop=vals_2,
            kwargs=kwargs,
        )
        printer.heading("Sweep2d: Sweep %s" % name_2)
        sweep_configs_2 = cls.sweep_1d_loop(
            args=args,
            name_sweep=name_2,
            vals_sweep=vals_2,
            name_loop=name_1,
            vals_loop=vals_1,
            kwargs=kwargs,
        )
        printer.heading("Sweep2d: Display %s" % name_1)
        CompareSweeps.run(
            args=args,
            config=sweep_configs_1,
            xlabel=name_1,
            clabel=name_2,
            log_x=(name_1 in log_x),
        )
        printer.heading("Sweep2d: Display %s" % name_2)
        CompareSweeps.run(
            args=args,
            config=sweep_configs_2,
            xlabel=name_2,
            clabel=name_1,
            log_x=(name_2 in log_x),
        )

    @classmethod
    def sweep_1d_loop(
        cls,
        args:       cli.ParsedArgs,
        name_sweep: str,
        vals_sweep: list,
        name_loop:  str,
        vals_loop:  list,
        kwargs:     dict,
    ) -> list[dict[str, str]]:
        sweep_configs = []
        for v in vals_loop:
            args.update({name_loop: v})
            sweep = Sweeper(
                args=args,
                **kwargs,
                **args.get_kwargs(),
                params={name_sweep: vals_sweep},
            )
            sweep_configs.append(
                {
                    "series_name":  v,
                    "sweep_name":   sweep.name,
                    "param_name":   name_sweep,
                },
            )

        return sweep_configs

    @classmethod
    def include_arg(cls, arg: cli.Arg) -> bool:
        return True if (arg.name != "devices") else False

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("name_1",       type=str,   required=True),
            cli.Arg("name_2",       type=str,   required=True),
            cli.JsonArg("vals_1",   nargs="*",  required=True),
            cli.JsonArg("vals_2",   nargs="*",  required=True),
            *[
                arg
                for arg in Sweeper.get_cli_options()
                if  arg.name != "params"
            ],
        ]



import torch
from jutility import cli
from juml.models.base import Model
from juml.models.sequential import Sequential
from juml.models.embed import Embedder
from juml.models.pool import Pooler

class RzCnn(Sequential):
    def __init__(
        self,
        input_shape:        list[int],
        output_shape:       list[int],
        kernel_size:        int,
        model_dim:          int,
        expand_ratio:       float,
        num_stages:         int,
        blocks_per_stage:   int,
        stride:             int,
        embedder:           Embedder,
        pooler:             Pooler,
    ):
        self._init_sequential(embedder, pooler)
        self.embed.set_input_shape(input_shape)
        self.pool.set_shapes([model_dim, None, None], output_shape)

        layer = torch.nn.Conv2d(
            in_channels=self.embed.get_output_dim(-3),
            out_channels=model_dim,
            kernel_size=kernel_size,
            stride=stride,
            padding=0,
        )
        self.layers.append(layer)

        for _ in range(num_stages - 1):
            for _ in range(blocks_per_stage - 1):
                layer = ReZeroCnnLayer(model_dim, expand_ratio, kernel_size)
                self.layers.append(layer)

            layer = torch.nn.Conv2d(
                in_channels=model_dim,
                out_channels=model_dim,
                kernel_size=kernel_size,
                stride=stride,
                padding=0,
            )
            self.layers.append(layer)

        for _ in range(blocks_per_stage):
            layer = ReZeroCnnLayer(model_dim, expand_ratio, kernel_size)
            self.layers.append(layer)

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("kernel_size",      type=int,   default=5),
            cli.Arg("model_dim",        type=int,   default=64),
            cli.Arg("expand_ratio",     type=float, default=2.0, tag="x"),
            cli.Arg("num_stages",       type=int,   default=3),
            cli.Arg("blocks_per_stage", type=int,   default=2),
            cli.Arg("stride",           type=int,   default=2),
        ]

class ReZeroCnnLayer(Model):
    def __init__(
        self,
        model_dim:      int,
        expand_ratio:   float,
        kernel_size:    int,
    ):
        self._torch_module_init()
        expand_dim = int(model_dim * expand_ratio)
        padding = kernel_size // 2

        self.conv_1 = torch.nn.Conv2d(
            in_channels=model_dim,
            out_channels=model_dim,
            kernel_size=kernel_size,
            stride=1,
            padding=padding,
            groups=model_dim,
        )
        self.conv_2 = torch.nn.Conv2d(
            in_channels=model_dim,
            out_channels=expand_dim,
            kernel_size=1,
            stride=1,
            padding=0,
            groups=1,
        )
        self.conv_3 = torch.nn.Conv2d(
            in_channels=expand_dim,
            out_channels=model_dim,
            kernel_size=1,
            stride=1,
            padding=0,
            groups=1,
        )
        self.scale = torch.nn.Parameter(torch.tensor(0.0))

    def forward(self, x_nmhw: torch.Tensor) -> torch.Tensor:
        res_nmhw = self.conv_1.forward(x_nmhw)
        res_nehw = self.conv_2.forward(res_nmhw)
        res_nehw = torch.relu(res_nehw)
        res_nmhw = self.conv_3.forward(res_nehw)
        x_nmhw = x_nmhw + (self.scale * res_nmhw)
        return x_nmhw



import os
import torch
from jutility import cli, util, plotting
from juml.device import DeviceConfig
from juml.models.base import Model
from juml.datasets.base import Dataset
from juml.loss.base import Loss

class Trainer:
    def __init__(
        self,
        args:       cli.ParsedArgs,
        model:      Model,
        dataset:    Dataset,
        loss:       Loss,
        device_cfg: DeviceConfig,
        table:      util.Table,
        **kwargs,
    ):
        self.model      = model
        self.dataset    = dataset
        self.loss       = loss
        self.table      = table

        self.train(
            args=args,
            model=model,
            dataset=dataset,
            loss=loss,
            device_cfg=device_cfg,
            table=table,
            **kwargs,
        )

        self.save_results(args)

    def train(
        self,
        args:       cli.ParsedArgs,
        model:      Model,
        dataset:    Dataset,
        loss:       Loss,
        device_cfg: DeviceConfig,
        table:      util.Table,
        **kwargs,
    ):
        raise NotImplementedError()

    @classmethod
    def from_args(
        cls,
        args:           cli.ParsedArgs,
        seed:           int,
        devices:        list[int],
        configs:        list[str],
        printer:        util.Printer,
        print_level:    int,
    ) -> "Trainer":
        cls.apply_configs(args, configs, [])
        torch.manual_seed(seed)

        dataset = cls.init_dataset(args)
        model   = cls.init_model(args, dataset)
        loss    = cls.init_loss(args, dataset)

        device_cfg = DeviceConfig(devices)
        device_cfg.set_module_device(model)
        loss.set_device(device_cfg)

        trainer_type = args.get_type("trainer")
        assert issubclass(trainer_type, Trainer)
        trainer_type.init_sub_objects(args, model, dataset)

        trainer = args.init_object(
            "trainer",
            args=args,
            model=model,
            dataset=dataset,
            loss=loss,
            device_cfg=device_cfg,
            table=util.Table(
                *trainer_type.get_table_columns(),
                printer=printer,
                print_interval=util.TimeInterval(1),
                print_level=print_level,
            ),
        )
        assert isinstance(trainer, Trainer)

        return trainer

    @classmethod
    def apply_configs(
        cls,
        args:           cli.ParsedArgs,
        config_paths:   list[str],
        forbidden:      list[str],
    ):
        for cp in config_paths:
            print("Loading config from \"%s\"" % cp)
            config_dict = util.load_json(cp)
            assert isinstance(config_dict, dict)

            extra_keys = set(config_dict.keys()) & set(forbidden)
            if len(extra_keys) > 0:
                raise ValueError(
                    "Configuration file \"%s\" contains forbidden keys %s"
                    % (cp, extra_keys)
                )

            args.update(config_dict)

    @classmethod
    def init_dataset(
        cls,
        args: cli.ParsedArgs,
    ) -> Dataset:
        with cli.verbose:
            dataset = args.init_object("dataset")
            assert isinstance(dataset, Dataset)

        return dataset

    @classmethod
    def init_model(
        cls,
        args:       cli.ParsedArgs,
        dataset:    Dataset,
    ) -> Model:
        with cli.verbose:
            model = args.init_object(
                "model",
                input_shape=dataset.get_input_shape(),
                output_shape=dataset.get_output_shape(),
            )
            assert isinstance(model, Model)

        return model

    @classmethod
    def init_loss(
        cls,
        args:       cli.ParsedArgs,
        dataset:    Dataset,
    ) -> Loss:
        loss_arg = args.get_arg("loss")
        loss_arg.set_default_choice(dataset.get_default_loss())

        with cli.verbose:
            loss = args.init_object("loss")
            assert isinstance(loss, Loss)

        if loss.needs_weights():
            loss.set_weights(dataset.get_loss_weights())

        return loss

    @classmethod
    def init_sub_objects(
        cls,
        args:       cli.ParsedArgs,
        model:      Model,
        dataset:    Dataset,
    ):
        raise NotImplementedError()

    def save_results(self, args: cli.ParsedArgs):
        self.output_dir = self.get_output_dir(args)
        self.model_name = os.path.basename(self.output_dir)

        time_list       = self.table.get_data("t")
        batch_loss      = self.table.get_data("batch_loss")
        train_metric    = self.table.get_data("train_metric")
        test_metric     = self.table.get_data("test_metric")

        cmd         = util.get_argv_str()
        arg_dict    = args.get_value_dict()
        metrics     = {
            "repr_model":   repr(self.model),
            "model_name":   self.model_name,
            "model_dir":    self.output_dir,
            "num_params":   self.model.num_params(),
            "time_str":     util.time_format(time_list[-1]),
            "time":         time_list[-1],
            "train":        {
                "start":    train_metric[0],
                "end":      train_metric[-1],
                "max":      max(train_metric),
                "min":      min(train_metric),
            },
            "test":         {
                "start":    test_metric[0],
                "end":      test_metric[-1],
                "max":      max(test_metric),
                "min":      min(test_metric),
            },
        }
        kw = {
            "item_fmt":     "%s = %8.5f",
            "key_order":    ["start", "end", "max", "min"],
        }
        metrics["train_summary"] = util.format_dict(metrics["train"], **kw)
        metrics[ "test_summary"] = util.format_dict(metrics["test" ], **kw)

        util.save_text(cmd,         "cmd",      self.output_dir)
        util.save_json(arg_dict,    "args",     self.output_dir)
        util.save_json(metrics,     "metrics",  self.output_dir)
        self.plot_metrics(batch_loss, train_metric, test_metric)
        self.save_model()
        self.save_table()
        print("--model_name %s" % self.model_name)
        print(
            "Final metrics = %.5f (train), %.5f (test)"
            % (train_metric[-1], test_metric[-1])
        )

    @classmethod
    def get_output_dir(cls, args: cli.ParsedArgs) -> str:
        return os.path.join("results", "train", cls.get_model_name(args))

    @classmethod
    def get_model_name(cls, args: cli.ParsedArgs) -> str:
        model_name = args.get_value("model_name")
        if model_name is None:
            model_name = cls.get_summary(args)

        return model_name

    @classmethod
    def get_summary(cls, args: cli.ParsedArgs) -> str:
        return "d%s_l%s_m%s_t%s_s%s" % tuple(
            a.get_value_summary() + a.get_summary()
            for a in [
                args.get_arg(name)
                for name in ["dataset", "loss", "model", "trainer", "seed"]
            ]
        )

    def plot_metrics(
        self,
        batch_loss:     list[float],
        train_metric:   list[float],
        test_metric:    list[float],
    ):
        train_label = "Train (final = %.5f)" % train_metric[-1]
        test_label  =  "Test (final = %.5f)" %  test_metric[-1]

        mp = plotting.MultiPlot(
            plotting.Subplot(
                plotting.Line(batch_loss),
                xlabel="Batch",
                **self.loss.info(),
            ),
            plotting.Subplot(
                plotting.Line(train_metric, c="b", label=train_label),
                plotting.Line(test_metric,  c="r", label=test_label),
                plotting.Legend(),
                xlabel="Epoch",
                **self.loss.metric_info(),
            ),
            title="%r\n%r" % (self.model, self.dataset),
            title_font_size=15,
            figsize=[10, 4],
        )
        mp.save("metrics", self.output_dir)

    def save_model(self):
        full_path = util.get_full_path("model.pth", self.output_dir)
        torch.save(self.model.state_dict(), full_path)

    def save_table(self):
        self.table.save_pickle("table", self.output_dir)

    @classmethod
    def load(cls, args: cli.ParsedArgs) -> tuple[str, Model, Dataset]:
        model_dir = cls.get_output_dir(args)

        args_path = util.get_full_path("args.json", model_dir, loading=True)
        args_dict = util.load_json(args_path)
        args.update(args_dict, allow_new_keys=True)

        torch.manual_seed(args.get_value("seed"))

        dataset = cls.init_dataset(args)
        model   = cls.init_model(args, dataset)

        model_path = util.get_full_path("model.pth", model_dir, loading=True)
        state_dict = torch.load(model_path, map_location="cpu")
        model.load_state_dict(state_dict)

        return model_dir, model, dataset

    @classmethod
    def get_metrics_path(cls, args: cli.ParsedArgs) -> str:
        return os.path.join(cls.get_output_dir(args), "metrics.json")

    @classmethod
    def get_table_columns(cls) -> list[util.Column]:
        raise NotImplementedError()

    @classmethod
    def get_loss_type(cls, args: cli.ParsedArgs) -> type[Loss]:
        dataset_type = args.get_type("dataset")
        assert issubclass(dataset_type, Dataset)

        loss_arg = args.get_arg("loss")
        loss_arg.set_default_choice(dataset_type.get_default_loss())
        loss_type = args.get_type("loss")
        assert issubclass(loss_type, Loss)

        return loss_type

    @classmethod
    def get_cli_arg(cls) -> cli.ObjectArg:
        return cli.ObjectArg(
            cls,
            *cls.get_cli_options(),
            tag=cls.get_tag(),
        )

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        raise NotImplementedError()

    @classmethod
    def get_tag(cls) -> (str | None):
        return None

    def __repr__(self):
        return util.format_type(type(self))



from juml.models import embed, pool

from juml.models.base import Model
from juml.models.sequential import Sequential
from juml.models.embed import Embedder
from juml.models.pool import Pooler
from juml.models.linear import Linear
from juml.models.linearmodel import LinearModel
from juml.models.mlp import Mlp
from juml.models.rzmlp import RzMlp, ReZeroMlpLayer
from juml.models.cnn import Cnn
from juml.models.rzcnn import RzCnn

def get_all() -> list[type[Model]]:
    return [
        LinearModel,
        Mlp,
        RzMlp,
        Cnn,
        RzCnn,
    ]



import torch
from jutility import cli
from juml.datasets.synthetic import Synthetic
from juml.loss.base import Loss
from juml.loss.mse import Mse

class SinMix(Synthetic):
    def __init__(
        self,
        input_dim:  int,
        hidden_dim: int,
        output_dim: int,
        train:      int,
        test:       int,
        x_std:      float,
        t_std:      float,
    ):
        self.w1_ih = torch.normal(0, 1, [input_dim,  hidden_dim])
        self.w2_ho = torch.normal(0, 1, [hidden_dim, output_dim])
        self.b1_h  = torch.normal(0, 1, [hidden_dim])
        self.b2_o  = torch.normal(0, 1, [output_dim])

        self._init_synthetic(
            input_shape=[input_dim],
            output_shape=[output_dim],
            n_train=train,
            n_test=test,
            x_std=x_std,
            t_std=t_std,
        )

    def _compute_target(self, x_ni: torch.Tensor) -> torch.Tensor:
        z_nh = x_ni @ self.w1_ih + self.b1_h
        z_nh = torch.sin(z_nh)
        t_no = z_nh @ self.w2_ho + self.b2_o
        return t_no

    @classmethod
    def get_default_loss(cls) -> type[Loss] | None:
        return Mse

    @classmethod
    def get_cli_arg(cls):
        return cli.ObjectArg(
            cls,
            cli.Arg("input_dim",    type=int,   default=1),
            cli.Arg("hidden_dim",   type=int,   default=1),
            cli.Arg("output_dim",   type=int,   default=1),
            cli.Arg("train",        type=int,   default=200),
            cli.Arg("test",         type=int,   default=200),
            cli.Arg("x_std",        type=float, default=0.0),
            cli.Arg("t_std",        type=float, default=0.1),
        )



from juml.commands.base import Command
from juml.commands.train import Train
from juml.commands.sweep import Sweep
from juml.commands.sweep2d import Sweep2d
from juml.commands.profile import Profile
from juml.commands.plot_confusion_matrix import PlotConfusionMatrix
from juml.commands.plot_1d_regression import Plot1dRegression
from juml.commands.plot_sequential import PlotSequential
from juml.commands.compare_sweeps import CompareSweeps

def get_all() -> list[type[Command]]:
    return [
        Train,
        Sweep,
        Sweep2d,
        Profile,
        PlotConfusionMatrix,
        Plot1dRegression,
        PlotSequential,
        CompareSweeps,
    ]



import torch
from jutility import cli, util
from juml.device import DeviceConfig
from juml.train.base import Trainer
from juml.train.bpsp import BpSp
from juml.models.base import Model
from juml.datasets.base import Dataset
from juml.loss.base import Loss

class BpSpDe(Trainer):
    def train(
        self,
        args:       cli.ParsedArgs,
        model:      Model,
        dataset:    Dataset,
        loss:       Loss,
        device_cfg: DeviceConfig,
        table:      util.Table,
        batch_size: int,
        steps:      int,
        n_train:    int,
        optimiser:  torch.optim.Optimizer,
        lrs:        torch.optim.lr_scheduler.LRScheduler,
    ):
        train_loader = dataset.get_subset_loader(
            split="train",
            subset_size=n_train,
            batch_size=batch_size,
        )
        test_loader  = dataset.get_data_loader(
            split="test",
            batch_size=batch_size,
        )

        table.get_column("train_metric").set_callback(
            callback=(lambda: loss.metric(model, train_loader, device_cfg)),
            level=1,
        )
        table.get_column("test_metric").set_callback(
            callback=(lambda: loss.metric(model, test_loader, device_cfg)),
            level=1,
        )

        s = 0
        e = 0
        table.update(level=1, step=s, epoch=e)
        while s < steps:
            for i, (x, t) in enumerate(train_loader):
                x, t = device_cfg.to_device([x, t])
                y = model.forward(x)
                batch_loss = loss.forward(y, t)

                optimiser.zero_grad()
                batch_loss.backward()
                optimiser.step()
                lrs.step()

                table.update(
                    step=s,
                    epoch=e,
                    batch=i,
                    batch_loss=batch_loss.item(),
                )
                s += 1
                if s >= steps:
                    table.update(level=1, step=s, epoch=e)
                    return

            e += 1

    @classmethod
    def init_sub_objects(
        cls,
        args:       cli.ParsedArgs,
        model:      Model,
        dataset:    Dataset,
    ):
        with cli.verbose:
            optimiser = args.init_object(
                "trainer.BpSpDe.optimiser",
                params=model.parameters(),
            )
            assert isinstance(optimiser, torch.optim.Optimizer)

        scheduler = args.init_object(
            "trainer.BpSpDe.lrs",
            optimizer=optimiser,
            T_max=args.get_value("trainer.BpSpDe.steps"),
        )
        assert isinstance(scheduler, torch.optim.lr_scheduler.LRScheduler)

    @classmethod
    def get_table_columns(cls) -> list[util.Column]:
        return [
            util.TimeColumn("t"),
            util.Column("step"),
            util.Column("epoch"),
            util.Column("batch"),
            util.Column("batch_loss", ".5f", width=10),
            util.CallbackColumn("train_metric", ".5f", width=12),
            util.CallbackColumn("test_metric",  ".5f", width=12),
        ]

    def save_model(self):
        return

    def save_table(self):
        return

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("steps",    type=int, default=int(1e5)),
            cli.Arg("n_train",  type=int, default=int(1e3)),
            *[arg for arg in BpSp.get_cli_options() if arg.name != "epochs"],
        ]

    @classmethod
    def get_tag(cls) -> (str | None):
        return "BSD"



import torch
from jutility import cli
from juml.models.base import Model
from juml.models.sequential import Sequential
from juml.models.embed import Embedder
from juml.models.pool import Pooler
from juml.models.linear import Linear

class RzMlp(Sequential):
    def __init__(
        self,
        input_shape:    list[int],
        output_shape:   list[int],
        model_dim:      int,
        expand_ratio:   float,
        depth:          int,
        embedder:       Embedder,
        pooler:         Pooler,
    ):
        self._init_sequential(embedder, pooler)
        self.embed.set_input_shape(input_shape)
        self.pool.set_shapes([model_dim], output_shape)

        layer = Linear(self.embed.get_output_dim(-1), model_dim)
        self.layers.append(layer)

        for _ in range(depth):
            layer = ReZeroMlpLayer(model_dim, expand_ratio)
            self.layers.append(layer)

        layer = Linear(model_dim, self.pool.get_input_dim(-1))
        self.layers.append(layer)

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("model_dim",    type=int,   default=100),
            cli.Arg("expand_ratio", type=float, default=2.0, tag="x"),
            cli.Arg("depth",        type=int,   default=3),
        ]

class ReZeroMlpLayer(Model):
    def __init__(
        self,
        model_dim:      int,
        expand_ratio:   float,
    ):
        self._torch_module_init()
        hidden_dim  = int(model_dim * expand_ratio)
        self.f1     = Linear(model_dim, hidden_dim)
        self.f2     = Linear(hidden_dim, model_dim)
        self.scale  = torch.nn.Parameter(torch.tensor(0.0))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x_res = self.f1.forward(x)
        x_res = torch.relu(x_res)
        x_res = self.f2.forward(x_res)
        x = x + (self.scale * x_res)
        return x



import torch
from jutility import cli, util
from juml.device import DeviceConfig
from juml.train.base import Trainer
from juml.models.base import Model
from juml.datasets.base import Dataset
from juml.loss.base import Loss

class BpSp(Trainer):
    def train(
        self,
        args:       cli.ParsedArgs,
        model:      Model,
        dataset:    Dataset,
        loss:       Loss,
        device_cfg: DeviceConfig,
        table:      util.Table,
        batch_size: int,
        epochs:     int,
        optimiser:  torch.optim.Optimizer,
        lrs:        torch.optim.lr_scheduler.LRScheduler,
    ):
        train_loader = dataset.get_data_loader("train", batch_size)
        test_loader  = dataset.get_data_loader("test" , batch_size)

        table.get_column("train_metric").set_callback(
            lambda: loss.metric(model, train_loader, device_cfg),
            level=1,
        )
        table.get_column("test_metric").set_callback(
            lambda: loss.metric(model, test_loader, device_cfg),
            level=1,
        )

        for e in range(epochs):
            table.update(level=1, epoch=e)
            for i, (x, t) in enumerate(train_loader):
                x, t = device_cfg.to_device([x, t])
                y = model.forward(x)
                batch_loss = loss.forward(y, t)

                optimiser.zero_grad()
                batch_loss.backward()
                optimiser.step()

                table.update(epoch=e, batch=i, batch_loss=batch_loss.item())

            table.print_last()
            lrs.step()

        table.update(level=2, epoch=epochs)

    @classmethod
    def init_sub_objects(
        cls,
        args:       cli.ParsedArgs,
        model:      Model,
        dataset:    Dataset,
    ):
        with cli.verbose:
            optimiser = args.init_object(
                "trainer.BpSp.optimiser",
                params=model.parameters(),
            )
            assert isinstance(optimiser, torch.optim.Optimizer)

        scheduler = args.init_object(
            "trainer.BpSp.lrs",
            optimizer=optimiser,
            T_max=args.get_value("trainer.BpSp.epochs"),
        )
        assert isinstance(scheduler, torch.optim.lr_scheduler.LRScheduler)

    @classmethod
    def get_table_columns(cls) -> list[util.Column]:
        return [
            util.TimeColumn("t"),
            util.Column("epoch"),
            util.Column("batch"),
            util.Column("batch_loss", ".5f", width=10),
            util.CallbackColumn("train_metric", ".5f", width=12),
            util.CallbackColumn("test_metric",  ".5f", width=12),
        ]

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("batch_size",   type=int, default=100),
            cli.Arg("epochs",       type=int, default=1),
            cli.ObjectChoice(
                "optimiser",
                cli.ObjectArg(
                    torch.optim.Adam,
                    cli.Arg("lr", type=float, default=0.001),
                ),
                cli.ObjectArg(
                    torch.optim.AdamW,
                    cli.Arg("lr",           type=float, default=0.001),
                    cli.Arg("weight_decay", type=float, default=0.01),
                    tag="AW",
                ),
                cli.ObjectArg(
                    torch.optim.SGD,
                    cli.Arg("lr",           type=float, default=0.001),
                    cli.Arg("momentum",     type=float, default=0.0),
                    cli.Arg("weight_decay", type=float, default=0.0),
                ),
                default="Adam",
            ),
            cli.ObjectChoice(
                "lrs",
                cli.ObjectArg(
                    torch.optim.lr_scheduler.CosineAnnealingLR,
                    cli.Arg("eta_min", type=float, default=1e-5),
                ),
                default="CosineAnnealingLR",
            ),
        ]



import torch
from jutility import cli
from juml.models.base import Model
from juml.models.sequential import Sequential
from juml.models.embed import Embedder
from juml.models.pool import Pooler
from juml.models.linear import Linear

class Mlp(Sequential):
    def __init__(
        self,
        input_shape:    list[int],
        output_shape:   list[int],
        hidden_dim:     int,
        depth:          int,
        embedder:       Embedder,
        pooler:         Pooler,
    ):
        self._init_sequential(embedder, pooler)
        self.embed.set_input_shape(input_shape)
        self.pool.set_shapes([], output_shape)

        layer_input_dim = self.embed.get_output_dim(-1)
        for _ in range(depth):
            layer = ReluMlpLayer(layer_input_dim, hidden_dim)
            self.layers.append(layer)
            layer_input_dim = hidden_dim

        layer_output_dim = self.pool.get_input_dim(-1)
        layer = Linear(layer_input_dim, layer_output_dim)
        self.layers.append(layer)

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return [
            cli.Arg("hidden_dim",   type=int, default=100),
            cli.Arg("depth",        type=int, default=3),
        ]

class ReluMlpLayer(Model):
    def __init__(
        self,
        input_dim:  int,
        output_dim: int,
    ):
        self._torch_module_init()
        self.linear = Linear(input_dim, output_dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.linear.forward(x)
        x = torch.relu(x)
        return x



import torch
from jutility import util, cli, units
from juml.models import embed, pool
from juml.models.base  import Model
from juml.models.embed import Embedder
from juml.models.pool import Pooler

class Sequential(Model):
    def __init__(
        self,
        embedder:   Embedder,
        layer_list: list[Model],
        pooler:     Pooler,
    ):
        self._init_sequential(embedder, pooler)
        for layer in layer_list:
            self.layers.append(layer)

    def _init_sequential(
        self,
        embedder:   Embedder,
        pooler:     Pooler,
    ):
        self._torch_module_init()
        self.embed  = embedder
        self.layers = torch.nn.ModuleList()
        self.pool   = pooler

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.embed.forward(x)

        for layer in self.layers:
            x = layer.forward(x)

        x = self.pool.forward(x)
        return x

    def split(self, layer_ind: int) -> tuple["Sequential", "Sequential"]:
        prefix = Sequential(
            self.embed,
            self.layers[:layer_ind],
            pool.Identity(),
        )
        suffix = Sequential(
            embed.Identity(),
            self.layers[layer_ind:],
            self.pool,
        )
        return prefix, suffix

    @classmethod
    def get_cli_arg(cls):
        return cli.ObjectArg(
            cls,
            *cls.get_cli_options(),
            embed.get_cli_choice(),
            pool.get_cli_choice(),
            tag=cls.get_tag(),
        )

    def __len__(self) -> int:
        return len(self.layers)



from jutility import cli, util
from juml import models, datasets, loss, train, commands
from juml.models.base import Model
from juml.datasets.base import Dataset
from juml.loss.base import Loss
from juml.train.base import Trainer
from juml.commands.base import Command

class Framework:
    @classmethod
    def get_models(cls) -> list[type[Model]]:
        return models.get_all()

    @classmethod
    def get_datasets(cls) -> list[type[Dataset]]:
        return datasets.get_all()

    @classmethod
    def get_losses(cls) -> list[type[Loss]]:
        return loss.get_all()

    @classmethod
    def get_trainers(cls) -> list[type[Trainer]]:
        return train.get_all()

    @classmethod
    def get_commands(cls) -> list[type[Command]]:
        return commands.get_all()

    @classmethod
    def get_defaults(cls) -> dict[str, str | None]:
        return {
            "model":    None,
            "dataset":  None,
            "loss":     None,
            "trainer":  "BpSp",
        }

    @classmethod
    def get_train_args(cls) -> list[cli.Arg]:
        defaults = cls.get_defaults()
        return [
            cli.ObjectChoice(
                "loss",
                *[t.get_cli_arg() for t in cls.get_losses()],
                default=defaults["loss"],
                is_group=True,
                required=False,
            ),
            cli.ObjectChoice(
                "dataset",
                *[t.get_cli_arg() for t in cls.get_datasets()],
                default=defaults["dataset"],
                is_group=True,
            ),
            cli.ObjectChoice(
                "model",
                *[t.get_cli_arg() for t in cls.get_models()],
                default=defaults["model"],
                is_group=True,
            ),
            cli.ObjectChoice(
                "trainer",
                *[t.get_cli_arg() for t in cls.get_trainers()],
                default=defaults["trainer"],
                is_group=True,
            ),
            cli.Arg("seed",         type=int, default=0),
            cli.Arg("devices",      type=int, default=[],   nargs="*"),
            cli.Arg("configs",      type=str, default=[],   nargs="*"),
            cli.Arg("model_name",   type=str, default=None, is_kwarg=False),
            cli.Arg("print_level",  type=int, default=0),
        ]

    @classmethod
    def get_parser(cls) -> cli.Parser:
        return cli.Parser(
            sub_commands=cli.SubCommandGroup(
                *[
                    command_type.init_juml(cls.get_train_args())
                    for command_type in cls.get_commands()
                ],
            ),
        )

    @classmethod
    def run(cls, *parser_args, **parser_kwargs):
        parser  = cls.get_parser()
        args    = parser.parse_args(*parser_args, **parser_kwargs)
        command = args.get_command()
        kwargs  = args.get_arg(command.name).get_kwargs()

        with util.Timer(repr(command)):
            command.run(args, **kwargs)



from jutility import cli

class Command(cli.SubCommand):
    @classmethod
    def run(
        cls,
        args: cli.ParsedArgs,
        **kwargs,
    ):
        raise NotImplementedError()

    @classmethod
    def init_juml(cls, train_args: list[cli.Arg]):
        return cls(
            cls.get_name(),
            *[
                arg
                for arg in train_args
                if  cls.include_arg(arg)
            ],
            cli.ArgGroup(
                cls.get_name(),
                *cls.get_cli_options(),
            ),
        )

    @classmethod
    def get_name(cls) -> str:
        return cls.__name__.lower()

    @classmethod
    def include_arg(cls, arg: cli.Arg) -> bool:
        return True

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return []

    def __repr__(self) -> str:
        return self.get_name()



import torch
from torch.autograd.profiler_util import FunctionEventAvg
from jutility import util, units
from juml.models.base import Model
from juml.datasets.base import Dataset
from juml.device import DeviceConfig
from juml.train.base import Trainer

class Profiler:
    def __init__(
        self,
        model:          Model,
        dataset:        Dataset,
        model_dir:      (str | None),
        batch_size:     int,
        num_warmup:     int,
        num_profile:    int,
        devices:        list[int],
        name:           str,
        sort_by:        str,
    ):
        device_cfg = DeviceConfig(devices)
        device_cfg.set_module_device(model)

        train_loader = dataset.get_data_loader("train", batch_size)
        x, t = next(iter(train_loader))
        [x] = device_cfg.to_device([x])

        for _ in range(num_warmup):
            y = model.forward(x)

        activities = [torch.profiler.ProfilerActivity.CPU]
        if len(devices) > 0:
            activities.append(torch.profiler.ProfilerActivity.CUDA)

        profiler_kwargs = {
            "activities":       activities,
            "profile_memory":   True,
            "with_flops":       True,
        }
        with torch.profiler.profile(**profiler_kwargs) as prof:
            with torch.profiler.record_function("model.forward"):
                for _ in range(num_profile):
                    y = model.forward(x)

        self.ka = prof.key_averages()
        printer = util.Printer(name, dir_name=model_dir)
        printer(self.ka.table(sort_by=sort_by))

        self.cpu_total  = self.get_cpu_total(self.ka)
        self.cuda_total = self.get_cuda_total(self.ka)
        self.t_total    = self.cpu_total + self.cuda_total
        self.n_samples  = batch_size * num_profile
        self.t_sample   = self.t_total / self.n_samples
        self.throughput = 1 / self.t_sample
        self.flops      = self.get_flops_total(self.ka) / self.n_samples
        profile_dict    = {
            "t_total":          self.t_total,
            "t_sample":         self.t_sample,
            "throughput":       self.throughput,
            "flops":            self.flops,
            "n_samples":        self.n_samples,
            "n_samples_str":    units.metric.format(self.n_samples),
            "batch_size":       batch_size,
            "n_repeats":        num_profile,
            "t_total_str": (
                "%.5f s"
                % self.t_total
            ),
            "t_sample_str": (
                "%.5f ms/sample"
                % (self.t_sample * 1e3)
            ),
            "throughput_str": (
                "%s samples/second"
                % units.metric.format(self.throughput)
            ),
            "flops_str": (
                "%s FLOPS/sample"
                % units.metric.format(self.flops).upper()
            ),
        }
        util.save_json(profile_dict, name, model_dir)

        md = util.MarkdownPrinter(name, model_dir)
        table = util.Table.key_value(printer=md)
        table.update(k="Model", v="`%s`" % repr(model))
        for name, dict_key in [
            ("Time (total)",    "t_total_str"),
            ("Time (average)",  "t_sample_str"),
            ("Throughput",      "throughput_str"),
            ("FLOPS",           "flops_str"),
            ("Total #samples",  "n_samples_str"),
            ("Batch size",      "batch_size"),
            ("`#`repeats",      "n_repeats"),
        ]:
            table.update(k=name, v=profile_dict[dict_key])

    @classmethod
    def get_cpu_total(cls, event_list: list[FunctionEventAvg]) -> float:
        return sum(e.self_cpu_time_total for e in event_list) * 1e-6

    @classmethod
    def get_cuda_total(cls, event_list: list[FunctionEventAvg]) -> float:
        return sum(e.self_cuda_time_total for e in event_list) * 1e-6

    @classmethod
    def get_flops_total(cls, event_list: list[FunctionEventAvg]) -> float:
        return sum(e.flops for e in event_list)



import torch.utils.data
from jutility import cli, util, units
from juml.datasets.split import DataSplit
from juml.loss.base import Loss

class Dataset:
    def get_data_loader(
        self,
        split: str,
        batch_size: int,
        shuffle=True,
        **kwargs,
    ) -> torch.utils.data.DataLoader:
        data_split = self.get_data_split(split)
        return torch.utils.data.DataLoader(
            dataset=data_split,
            batch_size=batch_size,
            shuffle=shuffle,
            **kwargs,
        )

    def get_subset_loader(
        self,
        split: str,
        subset_size: int,
        batch_size: int,
        shuffle=True,
        **kwargs,
    ) -> torch.utils.data.DataLoader:
        full_data_loader = self.get_data_loader(
            split=split,
            batch_size=subset_size,
            shuffle=True,
            **kwargs,
        )
        x, t = next(iter(full_data_loader))
        data_split = DataSplit(x, t, subset_size)
        return torch.utils.data.DataLoader(
            dataset=data_split,
            batch_size=batch_size,
            shuffle=shuffle,
            **kwargs,
        )

    def get_data_split(self, split: str) -> torch.utils.data.Dataset:
        raise NotImplementedError()

    def get_input_shape(self) -> list[int]:
        raise NotImplementedError()

    def get_output_shape(self) -> list[int]:
        raise NotImplementedError()

    @classmethod
    def get_default_loss(cls) -> type[Loss] | None:
        return None

    def get_loss_weights(self) -> torch.Tensor:
        raise NotImplementedError()

    def get_split_names(self) -> list[str]:
        raise NotImplementedError()

    @classmethod
    def get_cli_arg(cls):
        return cli.ObjectArg(cls)

    def __repr__(self):
        split_names = self.get_split_names()
        return util.format_type(
            type(self),
            **{
                ("n_" + split): units.metric.format(
                    len(self.get_data_split(split)),
                )
                for split in split_names
            },
            item_fmt="%s=%s",
            key_order=[("n_" + split) for split in split_names],
        )



from jutility import cli, util
from juml.commands.base import Command
from juml.train.base import Trainer

class Train(Command):
    @classmethod
    def run(
        cls,
        args: cli.ParsedArgs,
    ):
        return Trainer.from_args(
            args=args,
            printer=util.Printer(),
            **args.get_kwargs(),
        )



from juml.tools.experiment import Experiment, ExperimentGroup
from juml.tools.sweeper import Sweeper
from juml.tools.profiler import Profiler
from juml.tools.plot_type import PlotType
from juml.tools.display import display_sequential, plot_sequential



import torch
from jutility import util, cli
import juml

OUTPUT_DIR = juml.test_utils.get_output_dir("test_cli_args")

def test_cli_args_linearmodel():
    printer = util.Printer("test_cli_args_linearmodel", dir_name=OUTPUT_DIR)
    juml.test_utils.set_torch_seed("test_cli_args_linearmodel")

    parser = cli.Parser(
        cli.ObjectChoice(
            "model",
            juml.models.LinearModel.get_cli_arg(),
            juml.models.Mlp.get_cli_arg(),
            is_group=True,
        ),
        cli.ObjectChoice(
            "dataset",
            juml.datasets.LinearDataset.get_cli_arg(),
            juml.datasets.Mnist.get_cli_arg(),
            is_group=True,
        ),
    )
    arg_str = (
        "--model LinearModel "
        "--dataset LinearDataset "
        "--dataset.LinearDataset.input_dim  13  "
        "--dataset.LinearDataset.output_dim 7   "
        "--dataset.LinearDataset.train      789 "
        "--dataset.LinearDataset.test       456 "
    )
    args = parser.parse_args(arg_str.split())

    assert args.get_summary() == (
        "dLdi13do7dte456dtr789dts0.0dx0.0mLmeImpI"
    )

    cli.verbose.set_printer(printer)
    with cli.verbose:
        dataset = args.init_object("dataset")
        assert isinstance(dataset, juml.datasets.LinearDataset)

        model = args.init_object(
            "model",
            input_shape=dataset.get_input_shape(),
            output_shape=dataset.get_output_shape(),
        )
        assert isinstance(model, juml.models.LinearModel)

    data_loader = dataset.get_data_loader("train", 67)
    x, t = next(iter(data_loader))
    y = model.forward(x)

    assert repr(dataset) == "LinearDataset(n_train=789, n_test=456)"
    assert repr(model)   == "LinearModel(num_params=98)"

    assert dataset.get_input_shape()  == [13]
    assert dataset.get_output_shape() == [7]

    assert isinstance(x, torch.Tensor)
    assert isinstance(t, torch.Tensor)
    assert isinstance(y, torch.Tensor)

    assert list(x.shape) == [67, 13]
    assert list(t.shape) == [67, 7]
    assert list(y.shape) == [67, 7]

    printer.hline()
    printer(parser.help())
    printer(dataset)
    printer(model)

def test_cli_args_cnn():
    printer = util.Printer("test_cli_args_cnn", dir_name=OUTPUT_DIR)
    juml.test_utils.set_torch_seed("test_cli_args_cnn")

    parser = cli.Parser(
        cli.ObjectChoice(
            "model",
            juml.models.LinearModel.get_cli_arg(),
            juml.models.Mlp.get_cli_arg(),
            juml.models.Cnn.get_cli_arg(),
            is_group=True,
        ),
        cli.ObjectChoice(
            "dataset",
            juml.datasets.LinearDataset.get_cli_arg(),
            juml.datasets.Mnist.get_cli_arg(),
            is_group=True,
        ),
    )
    arg_str = (
        "--model Cnn "
        "--model.Cnn.pooler Average2d "
        "--model.Cnn.num_stages 2 "
        "--model.Cnn.blocks_per_stage 3 "
        "--dataset Mnist "
    )
    args = parser.parse_args(arg_str.split())

    assert args.get_summary() == (
        "dMmCmb3mc64meImk5mn2mpAms2"
    )

    cli.verbose.set_printer(printer)
    with cli.verbose:
        dataset = args.init_object("dataset")
        assert isinstance(dataset, juml.base.Dataset)

        model = args.init_object(
            "model",
            input_shape=dataset.get_input_shape(),
            output_shape=dataset.get_output_shape(),
        )
        assert isinstance(model, juml.base.Model)

    assert repr(dataset) == "Mnist(n_train=60.0k, n_test=10.0k)"
    assert repr(model)   == "Cnn(num_params=617.1k)"

    data_loader = dataset.get_data_loader("train", 67)
    x, t = next(iter(data_loader))
    y = model.forward(x)
    assert list(y.shape) == [67, 10]

    printer.hline()
    printer(parser.help())
    printer(dataset)
    printer(model)



from jutility import cli
from juml.commands.base import Command
from juml.tools.sweeper import Sweeper

class Sweep(Command):
    @classmethod
    def run(
        cls,
        args: cli.ParsedArgs,
        **kwargs,
    ):
        return Sweeper(
            args=args,
            **kwargs,
            **args.get_kwargs(),
        )

    @classmethod
    def include_arg(cls, arg: cli.Arg) -> bool:
        return True if (arg.name != "devices") else False

    @classmethod
    def get_cli_options(cls) -> list[cli.Arg]:
        return Sweeper.get_cli_options()



import torch
from jutility import cli, plotting
from juml.commands.base import Command
from juml.train.base import Trainer

class PlotConfusionMatrix(Command):
    @classmethod
    def run(
        cls,
        args: cli.ParsedArgs,
    ):
        model_dir, model, dataset = Trainer.load(args)

        test_split  = dataset.get_data_split("test")
        test_loader = dataset.get_data_loader("test", len(test_split))
        x, t = next(iter(test_loader))
        y = model.forward(x).argmax(dim=-1)

        c = torch.zeros([10, 10])
        for ground_truth in range(10):
            for prediction in range(10):
                n = (y[t == ground_truth] == prediction).sum()
                c[prediction, ground_truth] = n

        mp = plotting.MultiPlot(
            plotting.Subplot(
                plotting.ColourMesh(c),
                xlabel="Ground truth",
                ylabel="Prediction",
                axis_square=True,
            ),
            plotting.ColourBar(
                c.min(),
                c.max(),
                label="Number of classifications",
            ),
            width_ratios=[1, 0.1],
            figsize=[6, 5],
            title="Confusion matrix\n%r\n%r" % (model, dataset),
            title_font_size=15,
        )
        mp.save("Confusion matrix", model_dir)
